//
// Building cellular basis functions (textures) in realtime, based on code from ShaderX4
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#ifdef	MACOSX
	#include	<GLUT/glut.h>
#else
	#include	<glut.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>

#include    "libTexture.h"
#include    "TypeDefs.h"
#include    "Vector3D.h"
#include    "Vector2D.h"
#include	"utils.h"
#include	"fpTexture.h"
#include    "FrameBuffer.h"
#include    "GlslProgram.h"

#include	<list>

using namespace std;

#define	CELLSIZE	8
#define	DELTA		(12.0*16.0)

struct	Point
{
	Vector2D	pt;
	Vector2D	vel;
};

list<Point>	points;

GlslProgram prog1;
GlslProgram prog2;
GlslProgram prog3;
FrameBuffer	buf1 ( 512, 512, FrameBuffer :: depth32 );
FrameBuffer	buf2 ( 512, 512, FrameBuffer :: depth32 );
int	level = 1;

int	numPointsForCell ( float density )
{
	static	bool	inited = false;
	static	float	probs [10];
	int				m;
	
	if ( !inited )
	{
		float	lPower = 1;
		float	mFact  = 1;
		float	s      = 0;
		
		inited = true;
		
		for ( m = 0; m < 10; m++ )
		{
			float	p = lPower * exp ( -density ) / mFact;
			
			s        += p;
			mFact    *= m + 1;
			lPower   *= density;
			probs [m] = s;
		}
	}
	
	float	r = (float) rand () / (float) RAND_MAX;
	
	for ( m = 0; m < 10; m++ )
		if ( r < probs [m] )
			return m;
			
	return m;
}

void	genPoints ( float density = 2 )
{
	for ( int x = 0; x < CELLSIZE; x++ )
		for ( int y = 0; y < CELLSIZE; y++ )
		{
			int	numPoints = numPointsForCell ( density );
			
			for ( int i = 0; i < numPoints; i++ )
			{
				Vector2D pt ( Vector2D :: getRandomVector ( 0.5 ) );
				Vector2D v  ( Vector2D :: getRandomVector ( 0.05 ) );
				Point	 p;
				
				pt += Vector2D ( x, y ) / CELLSIZE;
				
				if ( pt.x < 0 )
					pt.x += 1;
					
				if ( pt.x > 1 )
					pt.x -= 1;
					
				if ( pt.y < 0 )
					pt.y += 1;
					
				if ( pt.y > 1 )
					pt.y -= 1;
					
				p.pt  = pt;
				p.vel = v;
				
				points.push_back ( p );
			}
		}
		
	printf ( "Total points: %d\n", points.size () );
}

void	fixCoords ()
{
	for ( list<Point>::iterator it = points.begin (); it != points.end (); ++it )
	{
		Vector2D pt ( it -> pt );
		
		if ( pt.x < 0 )
			pt.x += 1;
					
		if ( pt.x > 1 )
			pt.x -= 1;
					
		if ( pt.y < 0 )
			pt.y += 1;
					
		if ( pt.y > 1 )
			pt.y -= 1;
			
		it -> pt = pt;
	}
}

void	drawPoints ()
{
	glBegin ( GL_QUADS );
	
	for ( list<Point>::iterator it = points.begin (); it != points.end (); ++it )
	{
		Vector2D tex ( it -> pt	 );
		Vector2D pt  ( tex * Vector2D ( buf1.getWidth (), buf1.getHeight () ) );
		
		glTexCoord2fv ( tex );
		glVertex2f    ( pt.x - DELTA, pt.y - DELTA );
		glVertex2f    ( pt.x - DELTA, pt.y + DELTA );
		glVertex2f    ( pt.x + DELTA, pt.y + DELTA );
		glVertex2f    ( pt.x + DELTA, pt.y - DELTA );
	}
	
	glEnd ();
}

void	genF1 ()
{
	buf1.bind  ();
	prog1.bind ();
	
	glClear    ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	drawPoints ();
	
	prog1.unbind ();
	buf1.unbind  ();
}

void	genF2 ()
{
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf1.getColorBuffer () );
	
	buf2.bind  ();
	prog2.bind ();
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	drawPoints ();
	
	prog2.unbind ();
	buf2.unbind  ();
	
	glBindTexture ( GL_TEXTURE_2D, 0 );
}

void	genF3 ()
{
	glActiveTextureARB ( GL_TEXTURE0_ARB );
	glBindTexture      ( GL_TEXTURE_2D, buf2.getColorBuffer () );
	
	buf1.bind  ();
	prog3.bind ();
	
	glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	drawPoints ();
	
	prog3.unbind ();
	buf1.unbind  ();
	
	glBindTexture ( GL_TEXTURE_2D, 0 );
}

void	genCellularTexture ()
{
	startOrtho   ( buf1.getWidth (), buf1.getHeight () );
	glEnable     ( GL_DEPTH_TEST );
	glDepthFunc  ( GL_LEQUAL );
	glDepthMask  ( GL_TRUE );
	
	genF1        ();
	genF2        ();
	
	if ( level == 3 )
		genF3 ();
		
	endOrtho   ();
}

void display ()
{
    genCellularTexture ();

    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    startOrtho ( buf1.getWidth (), buf1.getHeight () );

	glEnable      ( GL_TEXTURE_2D );
	
	if ( level == 1 || level == 3 )
		glBindTexture ( GL_TEXTURE_2D, buf1.getColorBuffer () );
	else
	if ( level == 2 )
		glBindTexture ( GL_TEXTURE_2D, buf2.getColorBuffer () );

	drawQuad  ( buf1.getWidth (), buf1.getHeight () );
	
    glBindTexture ( GL_TEXTURE_2D, 0 );
		
    endOrtho ();

    glutSwapBuffers ();
}

void reshape ( int w, int h )		
{
   glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
   glMatrixMode   ( GL_PROJECTION );
   glLoadIdentity ();
   gluPerspective ( 60.0, (GLfloat)w/(GLfloat)h, 1.0, 60.0 );
   glMatrixMode   ( GL_MODELVIEW );
   glLoadIdentity ();
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
		
	if ( key == '1' )
		level = 1;
	else
	if ( key == '2' )
		level = 2;
	else
	if ( key == '3' )
		level = 3;
}

void    animate ()
{
	static	float	time = 0;
	float			t    = 0.001 * glutGet ( GLUT_ELAPSED_TIME );
	float			delta = t - time;
	
	time = t;
	
	for ( list<Point>::iterator it = points.begin (); it != points.end (); ++it )
		it -> pt += it -> vel * delta;
		
	fixCoords         ();
    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH );
    glutInitWindowSize  ( buf1.getWidth (), buf1.getHeight () );


                                // create window
    glutCreateWindow ( "GPU cellular textures" );

                                // register handlers
    glutDisplayFunc  ( display    );
    glutReshapeFunc  ( reshape    );
    glutKeyboardFunc ( key        );
    glutIdleFunc     ( animate    );

    init           ();
    initExtensions ();

	if ( !GlslProgram :: isSupported () )
	{
		printf ( "GLSL is not supported.\n" );
		
		return 1;
	}
	
	if ( !FrameBuffer :: isSupported () )
	{
		printf ( "Framebuffers not supported.\n" );
		
		return 1;
	}
	
    if ( !prog1.loadShaders ( "f1.vsh", "f1.fsh" ) )
    {
        printf ( "Error loading shaders f1:\n%s\n", prog1.getLog ().c_str () );

        return 3;
    }

    if ( !prog2.loadShaders ( "f1.vsh", "f2.fsh" ) )
    {
        printf ( "Error loading shaders f2:\n%s\n", prog2.getLog ().c_str () );

        return 3;
    }

    if ( !prog3.loadShaders ( "f1.vsh", "f3.fsh" ) )
    {
        printf ( "Error loading shaders f3:\n%s\n", prog3.getLog ().c_str () );

        return 3;
    }

	GLenum	fmt  = fpFormatWithPrecision ( 32 );
	GLenum	tex1 = buf1.createColorTexture ( GL_RGBA, fmt );
	GLenum	tex2 = buf2.createColorTexture ( GL_RGBA, fmt );
	
	buf1.create ();
	buf1.bind   ();
	buf1.attachColorTexture ( GL_TEXTURE_2D, tex1 );

	if ( !buf1.isOk () )
		printf ( "Error with framebuffer1\n" );

	buf1.unbind ();

	buf2.create ();
	buf2.bind   ();
	buf2.attachColorTexture ( GL_TEXTURE_2D, tex2 );

	if ( !buf2.isOk () )
		printf ( "Error with framebuffer2\n" );

	buf2.unbind ();
	
    prog2.bind            ();
    prog2.setTexture      ( "prevMap", 0 );
    prog2.unbind          ();

    prog3.bind            ();
    prog3.setTexture      ( "prevMap", 0 );
    prog3.unbind          ();

	genPoints    ();
    glutMainLoop ();

    return 0;
}
